/*	inline.h

	AFAPI TTL_PAPERS inline primitive functions....

	All magic constants have been written out explicitly to
	minimize potential name conflicts with user code.  Since it no
	longer damages readability ;-), we also have made a few
	optimizations which are dependent on the particular bit
	assignments within the port registers.  The other major change
	is that port addresses are now variables set by p_init(), thus
	allowing a single object code to run on machines with parallel
	ports at various different addresses.

	July 1996 by Hank Dietz
*/


#ifndef	_INLINE_H
#define	_INLINE_H


#include "afapi.h"

/*	p__inb() and p__outb()...

	These are the actual lowest-level hardware port I/O
	instructions used to touch the parallel port interface.  They
	are defined here, rather than borrowed from a library, to
	ensure that they don't add any delays...  many library
	versions insert code to allow port settle time, which we
	confirm using a handshake protocol instead.
*/

#ifdef	P_DEBOUNCE

extern inline
unsigned int
p___inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

extern inline
unsigned int
p__inb(unsigned short port)
{
	register unsigned int t = p___inb(port);

	while (t != p_inlast) {
		p_inlast = t;
		t = p___inb(port);
	}
	return(t);
}

#else

extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

#endif


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}

/*	p__wait()...

	This code is newly optimized for the 960801 release.  Old
	versions of p_wait() used an if-else to deal with the fact
	that the sense of P_RDY toggles; instead, here we make use of
	the fact that an appropriate value to match can be built
	simply by comparing P_RDY with the P_S1 we sent (both happen
	to be in the same bit position in their corresponding
	registers).

*/

extern inline
void
p__wait(void)
{
	register unsigned char t;
	register unsigned char match;

	/* Flip the strobe bits... */
	t = (p_outlast ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t, p_outport);
		}
	}
}

/*	p__nand()...

	This code uses the same optimizations found in the new
	p__wait().  Additionally, it takes advantage of the fact that
	flipping the strobe bits twice yields the original pattern.
	However, to maximize these benefits, it did switch from using
	early output to using late read for increasing data settle
	time to avoid races caused by badly-designed parallel ports.
*/	

extern inline
int
p__nand(register p_8u b)
{
	register unsigned char t, t2;
	register unsigned char match;
	register int result;

	/* Paste-in the new data, flip strobe bits, and output it */
	t2 = ((p_outlast & 0xf0) | (b & 0x0f));
	t = (t2 ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t2;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t, p_outport);
		}
	}

	/* Read the data value again (time for data to settle) */
	result = p__inb(p_inport);

	/* Now do another barrier to say everyone has seen the data
	   value.  Notice that we save a little overhead by reusing
	   the match value from above....
	*/
	p__outb(t2, p_outport);
	if ((p__inb(p_inport) & 0x80) != match) {
		if ((p__inb(p_inport) & 0x80) != match) {
			p__outb((t2 ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) != match) ;
			p__outb(t2, p_outport);
		}
	}

	/* Return with data bit vector shifted into place and masked
	   to 4 bits.
	*/
	return((result >> 3) & 0xf);
}

/*	p__put() and p__get()...

	This code uses the same optimizations found in the new
	p__nand(), but also takes advantage of the fact that the code
	doesn't have to be the same for sender and reciever.
*/	

extern inline
int
p__get(void)
{
	register unsigned char t, t2;
	register unsigned char match;
	register int result;

	/* Paste-in 0xf data, flip strobe bits, and output it */
	t2 = (p_outlast | 0x0f);
	t = (t2 ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t2;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t, p_outport);
		}
	}

	/* Read the data value again (time for data to settle) */
	result = p__inb(p_inport);

	/* Now do another barrier to say everyone has seen the data
	   value.  Notice that we save a little overhead by reusing
	   the match value from above....
	*/
	p__outb(t2, p_outport);
	if ((p__inb(p_inport) & 0x80) != match) {
		if ((p__inb(p_inport) & 0x80) != match) {
			p__outb((t2 ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) != match) ;
			p__outb(t2, p_outport);
		}
	}

	/* Return with data bit vector shifted into place and masked
	   to 4 bits.  Note the complement is needed to undo the NAND.
	*/
	return(((~result) >> 3) & 0xf);
}


extern inline
void
p__put(register p_8u b)
{
	register unsigned char t, t2;
	register unsigned char match;

	/* Paste-in the new data, flip strobe bits, and output it */
	t2 = ((p_outlast & 0xf0) | (b & 0x0f));
	t = (t2 ^ (0x80 | 0x40));
	p__outb(t, p_outport);
	p_outlast = t2;

	/* Wait for P_RDY state to match P_S state... this is done by
	   setting match to the bit we want to change, and then
	   comparing the masked input with match.  If we poll more
	   than twice, turn on the Red LED until the wait has been
	   satisfied.
	*/
	match = (t & 0x80);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t, p_outport);
		}
	}

	/* Skip the read of the new data value... we know it */

	/* Now do another barrier to say everyone has seen the data
	   value.  Notice that we save a little overhead by reusing
	   the match value from above....
	*/
	p__outb(t2, p_outport);
	if ((p__inb(p_inport) & 0x80) != match) {
		if ((p__inb(p_inport) & 0x80) != match) {
			p__outb((t2 ^ (0x20 | 0x10)), p_outport);
			while ((p__inb(p_inport) & 0x80) != match) ;
			p__outb(t2, p_outport);
		}
	}
}

/*	p__waitbar(), p__any(), and p_all()...

	These really are not bottom-level primitives for the current
	TTL_PAPERS hardware, but each is small enough to inline....
*/

extern inline
p_bar
p__waitbar(register p_1u f)
{
	/* Collect a barrier bit mask with the votes from all PEs. Loop
	   collecting highest bits first...  some magic is required to
	   ensure that non multiple-of-4 NPROC values work.
	*/
	register p_bar t = 0;
	register int i = ((NPROC - 1) & ~0x3);

	do {
		t = ((t << 4) |
		     p__nand((f && (IPROC >= i) && (IPROC < i+4)) ?
			     (0xf ^ (1 << (IPROC - i))) :
			     0xf));
	} while ((i -= 4) >= 0);

	/* Only need to mask if NPROC bits isn't the actual size of
	   the barrier (p_bar) data type.
	*/
#if	((NPROC == 8) || (NPROC == 16) || (NPROC == 32))
	return(t);
#else
	return(t & (((p_bar) (1 << NPROC)) - 1));
#endif
}


extern inline
p_1u
p__any(register p_1u f)
{
	return(p__nand(f ? 0xe : 0xf));
}


extern inline
p_1u
p__all(register p_1u f)
{
	return(!p__nand(f ? 0xf : 0xe));
}

/*	Inline signal primitives...

	Very simple stuff...  barrier NAND hardware does signals very
	naturally.
*/

extern inline void
s__signal(register int s)
{
	/* This is not worth inlining... */
	s_signal(s);
}


extern inline void
s__poll(void)
{
	/* Really easy when you got the right hardware ;-) */
	if (p__inb(p_inport) & 0x38) {
		s_signal(S_SIGANY);
	}
}


extern inline void
s__wait(void)
{
	/* Just exit and re-enter... that flushes things */
	s_exit();
	s_init();
}

/*	Inline RSM interface...

	It happens that only the signal handler itself is a lot of
	code, thus, these can also be inlined....
*/

extern inline void
s__update(register void *a,
register int n)
{
	s_updateArgs.a = a;
	s_updateArgs.n = n;
	s__signal(S_SIGMEM);
}


extern inline void
s__initLock(register s_lock *lock)
{
	/* Initialize lock... */
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


extern inline void
s__acquire(register s_lock *lock)
{
	/* Acquire lock...  the funky thing about this is that it
	   works for our system BECAUSE replicated shared memory
	   updates are internally executed in a purely synchronous
	   way...  it wouldn't work using the SMP hardware directly.
	*/
	register s_lock us = (IPROC + 1);

	do {
		/* Wait for lock to be open... */
		while (*lock != S_INITLOCK) {
			s__poll();
		}

		/* Try to grab it for us */
		*lock = us;
		s__update(((void *) lock), sizeof(s_lock));

		/* Continue looping if we lost the race... */
	} while (*lock != us);
}


extern inline void
s__release(register s_lock *lock)
{
	/* Unlock lock.
	   Hmm...  this looks like s__initLock(), doesn't it?
	*/
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


#endif
